#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdexcept>


#ifdef HAVE_IO_H
#include <io.h>
#endif
#ifdef O_BINARY
#define OUR_O_BINARY O_BINARY
#else
#define OUR_O_BINARY 0
#endif

#ifdef O_LARGEFILE
#define OUR_O_LARGEFILE O_LARGEFILE
#else
#define OUR_O_LARGEFILE 0
#endif




/***************************
*   erf.cpp
*   author:  Steve Strand
*   written: 29-Jan-04
***************************/
static const double rel_error= 1E-12;        //calculate 12 significant figures
//you can adjust rel_error to trade off between accuracy and speed
//but don't ask for > 15 figures (assuming usual 52 bit mantissa in a double)


double erf(double x)
//erf(x) = 2/sqrt(pi)*integral(exp(-t^2),t,0,x)
//       = 2/sqrt(pi)*[x - x^3/3 + x^5/5*2! - x^7/7*3! + ...]
//       = 1-erfc(x)
{
    static const double two_sqrtpi=  1.128379167095512574;        // 2/sqrt(pi)
    if (fabs(x) > 2.2) {
        return 1.0 - erfc(x);        //use continued fraction when fabs(x) > 2.2
    }
    double sum= x, term= x, xsqr= x*x;
    int j= 1;
    do {
        term*= xsqr/j;
        sum-= term/(2*j+1);
        ++j;
        term*= xsqr/j;
        sum+= term/(2*j+1);
        ++j;
    } while (fabs(term)/sum > rel_error);
    return two_sqrtpi*sum;
}


double erfc(double x)
//erfc(x) = 2/sqrt(pi)*integral(exp(-t^2),t,x,inf)
//        = exp(-x^2)/sqrt(pi) * [1/x+ (1/2)/x+ (2/2)/x+ (3/2)/x+ (4/2)/x+ ...]
//        = 1-erf(x)
//expression inside [] is a continued fraction so '+' means add to denominator only
{
    static const double one_sqrtpi=  0.564189583547756287;        // 1/sqrt(pi)
    if (fabs(x) < 2.2) {
        return 1.0 - erf(x);        //use series when fabs(x) < 2.2
    }
    if (signbit(x)) {               //continued fraction only valid for x>0
        return 2.0 - erfc(-x);
    }
    double a=1, b=x;                //last two convergent numerators
    double c=x, d=x*x+0.5;          //last two convergent denominators
    double q1, q2= b/d;             //last two convergents (a/c and b/d)
    double n= 1.0, t;
    do {
        t= a*n+b*x;
        a= b;
        b= t;
        t= c*n+d*x;
        c= d;
        d= t;
        n+= 0.5;
        q1= q2;
        q2= b/d;
      } while (fabs(q1-q2)/q2 > rel_error);
    return one_sqrtpi*exp(-x*x)*q2;
}


double snr2ber(int bps, double snr)
{
	double ber, ser;
	int m;
	
	snr = pow(10, snr / 10);	// dB to power
	snr *= 80;
	
	if(bps == 1) {
		m = 2;
		ber = erfc(pow(snr, 0.5) ) / 2;
	}
	else if(bps == 2) {
		m = 4;
		ber = erfc(pow(snr / 2, 0.5) ) / 2;
	}
        else if(bps == 3) {
                m = 8;
                ber = erfc(pow((9 * snr) / 7, 0.5) ) * 4;
        }

	else if(bps == 4) {
		m = 16;
		ber = erfc(pow(snr / 10, 0.5) ) * 3 / 8;
	}
	else if(bps == 6) {
		m = 64;
		ber = erfc(pow(snr / 42, 0.5) ) * 7 / 24;
	}
	
	return ber;
}

void readCSISubcarriers(float *csi, const int total_subcarriers)
{
   const char *filename1 = "csi.dat"; //"l1.dat_matrix_dB";
   printf("total_subcarriers: %d\n", total_subcarriers); fflush(stdout);
   int fd;
   FILE *d_fp;

   int count = 0;
   if ((fd = open (filename1,
                   O_RDONLY|OUR_O_LARGEFILE,
                   0664)) < 0){
        perror(filename1);
   }else {
        if((d_fp = fdopen (fd, true ? "r" : "rt")) == NULL) {
           fprintf(stderr, "CSI file cannot be opened\n");
           close(fd);
           assert(false);
        }
        char line[1700];

        float *tmp = (float*) malloc(sizeof(float) * total_subcarriers);
        memset(tmp, 0, sizeof(float) * total_subcarriers);

        int sc_index = 0;
        while(fgets(line, 1700, d_fp) != NULL)
        {
                sscanf(line, "%f", &csi[sc_index]);
                sc_index++;
        }
        fclose(d_fp);
        free(tmp);
   }
}

int main(int argc, char *argv[])
{
    int subcarriers = 192;
    int bps = 6;
    float *csi = (float*) malloc(sizeof(float) * subcarriers);
    memset(csi, 0, sizeof(float) * subcarriers);

    readCSISubcarriers(csi, subcarriers);

    for(int i = 0; i < subcarriers; i++)
    	printf("%lf\n", snr2ber(bps, csi[i]));
    free(csi); 
}
